#include <dlfcn.h>
#include <iostream>

#include <glib-2.0/glib.h>

#include "nvdsinfer_model_builder.h"

using namespace nvdsinfer;

class NvLogger : public nvinfer1::ILogger {
    void log(Severity severity, const char* msg) override {
        if (severity != Severity::kINFO and severity != Severity::kVERBOSE) {
            std::cout << "NvLogger: " << msg << std::endl;
        }
    }
};

void resetParams(NvDsInferContextInitParams& params) {
    memset(&params, 0, sizeof(params));

    params.networkMode = NvDsInferNetworkMode_FP32;
    params.networkInputFormat = NvDsInferFormat_Unknown;
    params.uffInputOrder = NvDsInferTensorOrder_kNCHW;
    params.maxBatchSize = 1;
    params.networkScaleFactor = 1.0;
    params.networkType = NvDsInferNetworkType_Detector;
    params.outputBufferPoolSize = NVDSINFER_MIN_OUTPUT_BUFFERPOOL_SIZE;
}

void getYoloV3Params(NvDsInferContextInitParams& params) {
    resetParams(params);

    params.uniqueID = 1;
    params.gpuID = 0;
    params.networkMode = NvDsInferNetworkMode_FP32;
    params.maxBatchSize = 1;

    std::string yolov3_path = "/opt/nvidia/deepstream/deepstream-5.0/sources/objectDetector_Yolo/";
    std::string model_path = yolov3_path + std::string("yolov3-tiny.weights");
    std::string cfg_path = yolov3_path + std::string("yolov3-tiny.cfg");
    std::string labels_path = yolov3_path + std::string("labels.txt");
    std::string lib_path = yolov3_path + std::string("nvdsinfer_custom_impl_Yolo/libnvdsinfer_custom_impl_Yolo.so");

    g_strlcpy(params.modelFilePath, model_path.c_str(), _PATH_MAX);
    g_strlcpy(params.customNetworkConfigFilePath, cfg_path.c_str(), _PATH_MAX);
    g_strlcpy(params.labelsFilePath, labels_path.c_str(), _PATH_MAX);
    g_strlcpy(params.customLibPath, lib_path.c_str(), _PATH_MAX);

    g_strlcpy(params.customEngineCreateFuncName, "NvDsInferYoloCudaEngineGet", _MAX_STR_LENGTH);
    g_strlcpy(params.customBBoxParseFuncName, "NvDsInferParseCustomYoloV3Tiny", _MAX_STR_LENGTH);

    params.numDetectedClasses = 80;
    params.clusterMode = NVDSINFER_CLUSTER_NMS;
    params.networkScaleFactor = 1.0 / 255;
    params.networkInputFormat = NvDsInferFormat_RGB;
    params.networkType = NvDsInferNetworkType_Detector;
}

void getPeopleNetParams(NvDsInferContextInitParams& params) {
    resetParams(params);

    params.uniqueID = 1;
    params.gpuID = 0;
    params.networkMode = NvDsInferNetworkMode_FP32;
    params.maxBatchSize = 1;

    std::string peoplenet_path = "/opt/nvidia/deepstream/deepstream-5.0/samples/models/tlt_pretrained_models/peoplenet/";
    std::string model_path = peoplenet_path + std::string("resnet34_peoplenet_pruned.etlt");
    std::string labels_path = peoplenet_path + std::string("labels_peoplenet.txt");
    g_strlcpy(params.tltEncodedModelFilePath, model_path.c_str(), _PATH_MAX);
    g_strlcpy(params.labelsFilePath, labels_path.c_str(), _PATH_MAX);

    g_strlcpy(params.tltModelKey, "tlt_encode", _MAX_STR_LENGTH);

    params.numDetectedClasses = 3;
    params.clusterMode = NVDSINFER_CLUSTER_DBSCAN;
    params.networkScaleFactor = 1.0 / 255;
    params.networkInputFormat = NvDsInferFormat_RGB;
    params.networkType = NvDsInferNetworkType_Detector;

    params.uffInputOrder = NvDsInferTensorOrder_kNCHW;
    params.inferInputDims = NvDsInferDimsCHW{3, 544, 960};
    g_strlcpy(params.uffInputBlobName, "input_1", _MAX_STR_LENGTH);
    params.outputLayerNames = g_strsplit("output_bbox/BiasAdd;output_cov/Sigmoid", ";", 2);
    params.numOutputLayers = 2;
}

void buildModel(const NvDsInferContextInitParams& params) {
    NvLogger logger;

    /* Load the custom library if specified. */
    std::shared_ptr<DlLibHandle> dlHandle;
    if (!string_empty(params.customLibPath)) {
        dlHandle = std::make_shared<DlLibHandle>(params.customLibPath, RTLD_LAZY);
        if (!dlHandle->isValid()) {
            throw std::runtime_error("Could not open custom lib: " + std::string(dlerror()));
        }
    }

    /* Instantiate Builder */
    auto builder = std::make_unique<TrtModelBuilder>(params.gpuID, logger, dlHandle);
    if (!builder) {
        throw std::runtime_error("Could not instantiate builder");
    }

    /* Build engine */
    std::string enginePath;  // suggested path
    std::shared_ptr<TrtEngine> engine = builder->buildModel(params, enginePath);
    if (!engine) {
        throw std::runtime_error("Build engine file failed");
    }
}

int main() {
    /* 1. Build yolo */
    std::cout << "=== YOLO ===" << std::endl;
    NvDsInferContextInitParams yolo_params;
    getYoloV3Params(yolo_params);
    buildModel(yolo_params);

    /* 2. Build Peoplenet */
    std::cout << "=== PEOPLENET ===" << std::endl;
    NvDsInferContextInitParams peoplenet_params;
    getPeopleNetParams(peoplenet_params);
    buildModel(peoplenet_params);

    return 0;
}
